/**
 *   \file		lcscom.c
 *   \author	Tranchida Giulio, No. Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.\n\n
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>      /* indirizzo AF_UNIX */

#include "Include/common.h"
#include "Include/lcscom.h"

/* Crea un channel di ascolto AF_UNIX */
serverChannel_t createServerChannel(const char* path){
	serverChannel_t server_sk;
	struct sockaddr_un servaddr;     /* address of server */
    
	/* Creo il socket */
	if ((server_sk = socket(AF_UNIX,SOCK_STREAM,0)) < 0){
		errore(__FILE__,__LINE__,"Server: socket create",errno);
		return -1;
	}
	if (strlen(path) > UNIX_PATH_MAX){
		errore(__FILE__,__LINE__,"Server: socket path too long (exceeding UNIX_PATH_MAX)",errno);    
		return SFATENAMETOOLONG;
	}    
	memset(&servaddr, 0, sizeof(struct sockaddr_un));	/* Puliamo la struttura */  
	servaddr.sun_family = AF_UNIX;
	strncpy(servaddr.sun_path, path, UNIX_PATH_MAX);
    
	/* Rendiamo il socket visibile nel dominio */
	if (bind(server_sk, (struct sockaddr *)&servaddr, sizeof(struct sockaddr_un)) < 0){
		close(server_sk);
		errore(__FILE__,__LINE__,"Server: socket bind",errno);
		return -1;     
	}

	/* Mettiamo il socket in attesa di connessioni */
	if (listen(server_sk,MAXCONN) < 0){
		close(server_sk);
		errore(__FILE__,__LINE__,"Server: socket listen",errno);		
		return -1;
	}
	return server_sk;
}

/* Chiude e distrugge una socket creata dal server */
int closeSocket(serverChannel_t s){
	/* Chiudo la socket in lettura e scrittura (invio SEOF al client) */
	if (shutdown(s, SHUT_RDWR) == -1){
		errore(__FILE__,__LINE__,"client: socket shutdown",errno);		
		return -1;	
	}	
	if (close(s)){
		errore(__FILE__,__LINE__,"Server: socket close",errno);		
		return -1;	
	}
	return 0;
}

/* Accetta una connessione da parte di un client e ritorna il
 * descrittore del channel di trasmissione con il client */
channel_t acceptConnection(serverChannel_t s){
	channel_t  sk_comunication;
	
	/* In caso di errore la funzione accept ritorna -1 */
	if ((sk_comunication = accept(s, NULL, 0)) == -1)
		errore(__FILE__,__LINE__,"Server: socket accept",errno);
	return sk_comunication;
}

/* Scrive un messaggio sul channel di trasmissione rispettando
 * la struttura che contiene il messaggio da scrivere  */ 
int sendMessage(channel_t sc, message_t *msg){
	int char_sent, sent;
	int nsec;
	
	/* write ritorna -1 in caso di errore, altrimenti il numero di caratteri scritti */
	/* Temporizzazione in caso di errore. (^l^) */
	for (nsec = 1; nsec<=MAXSLEEP; nsec<<=1){
		if ((sent = write(sc, &(msg->type), sizeof(char))) != -1) break;
		if (nsec<=MAXSLEEP/2) sleep(nsec);
	}
	if (sent==-1){ errore(__FILE__,__LINE__,"Error send message 'type'",errno); return -1; }
	char_sent = sent;
	
	for (nsec = 1; nsec<=MAXSLEEP; nsec<<=1){
		if ((sent = write(sc, &(msg->length), sizeof(unsigned int))) != -1) break;
		if (nsec<=MAXSLEEP/2) sleep(nsec);
	}
	if (sent==-1){ errore(__FILE__,__LINE__,"Error send message 'length'",errno); return -1; }
	char_sent += sent;

	if (msg->length!=0){
		for (nsec = 1; nsec<=MAXSLEEP; nsec<<=1){
			if ((sent = write(sc, msg->buffer, msg->length)) != -1) break;
			if (nsec<=MAXSLEEP/2) sleep(nsec);
		}
		if (sent==-1){ errore(__FILE__,__LINE__,"Error send message 'buffer'",errno); return -1; }		
		char_sent += sent;
	}
	if (msg->buffer) free(msg->buffer);
	msg->buffer = NULL;
	return char_sent;
}

/* Legge un messaggio dal channel di trasmissione */
int receiveMessage(channel_t sc, message_t *msg){
	int letti, read_cnt;
	if ((letti = read(sc, &(msg->type), sizeof(char))) < 0 ){
		errore(__FILE__,__LINE__,"Error read message 'type'",errno);
		return -1;		
	}else if (letti==0) return SEOF; /* EOF sul socket */
	read_cnt = letti;
	if ((letti = read(sc, &(msg->length), sizeof(unsigned int))) < 0){
		errore(__FILE__,__LINE__,"Error read message 'length'",errno);
		return -1;		
	}else if (letti==0) return SEOF; /* EOF sul socket */
	read_cnt += letti;
	
	if (msg->length!=0){
		/* alloco il buffer di di lenght + 1 al fine di terminare la stringa */
		if (!(msg->buffer = calloc(msg->length +1, sizeof(char)))){
			errore(__FILE__,__LINE__,"Error calloc 'buffer'",errno);
			return -1;			
		}
		if ((letti = read(sc, msg->buffer, msg->length)) <= 0){
			errore(__FILE__,__LINE__,"Error read message 'buffer'",errno);
			free(msg->buffer);
			msg->buffer = NULL;
			if (letti==0) return SEOF; /* EOF sul socket */
			return -1;
		}
		read_cnt += letti;
	}else
		msg->buffer = NULL;

	return read_cnt;
}

/* Chiude e distrugge una socket (client) */
int closeConnection(channel_t sc){
	/* Chiudo la socket lato client (invio SEOF al server) */
	if (shutdown(sc, SHUT_WR) == -1){
		errore(__FILE__,__LINE__,"client: socket shutdown",errno);		
		return -1;	
	}
	if (close(sc)){
		errore(__FILE__,__LINE__,"client: socket close",errno);		
		return -1;	
	}
	return 0;
}

/* Crea un channel di trasmissione verso il server */
channel_t openConnection(const char* path){
	channel_t client_sk;
	struct sockaddr_un clientaddr;     /* address of server */
    int nsec;
    
	/* Creo il socket */
	if ((client_sk = socket(AF_UNIX,SOCK_STREAM,0)) < 0){
		errore(__FILE__,__LINE__,"Client: socket create",errno);
		return -1;
	}
	if (strlen(path) > UNIX_PATH_MAX ){
		errore(__FILE__,__LINE__,"Client: socket path too long (exceeding UNIX_PATH_MAX)",errno);    
		return SFATENAMETOOLONG;
	}
	memset(&clientaddr, 0, sizeof(struct sockaddr_un));	/* Puliamo la struttura */  
	clientaddr.sun_family = AF_UNIX;
	strncpy(clientaddr.sun_path, path, UNIX_PATH_MAX);
    
	/* Connessione con riprova */
	for (nsec = 1; nsec<=MAXSLEEP; nsec<<=1){
		/* Richiedo la connessione alla socket del server */
		if (connect(client_sk, (struct sockaddr *)&clientaddr, sizeof(struct sockaddr_un)) == 0)
			return client_sk;
		if (nsec<=MAXSLEEP/2) sleep(nsec);
	}
	errore(__FILE__,__LINE__-3,"Client: socket connect",errno);
	return -1;
}


