/**
 *	\file		server.c
 *	\author		Tranchida Giulio, No Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.\n\n
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */

#include "Include/server.h"

/** Main / Dispather
 * - Registro il mio gestore e la mia maschera dei segnali,
 * - Controllo che il filesystem esista, che sia un file regolare e lo
 *   monto,
 * - Creo la socket,
 * - Entro in un ciclo infinito in cui riamango in attesa di una connessione
 *   da parte di un nuovo client,
 * - Una volta che un nuovo client ha stabilito una connessione, creo un
 *   thread che gestira' le comunicazioni tra quel client e il server
 * - La condizione di uscita dal ciclo e' quitflag != 0, impostato ad 1,
 *   esclusivamente nel gestore dei segnali.
 * - Una volta uscito dal ciclo, attendo la chiusura di ogni thread attivo,
 * - Chiudo la socket e la elimino fisicamente dal filesystem.
 * - Libero le ultime risorse allocate ed esco.
 * */
int main(int argc, char **argv) {
	extern FILE * fs;
	extern struct fat_ctrl f_ctrl;
	extern int num_threads;
	extern pthread_mutex_t lock;
	extern serverChannel_t sk_ascolto;
	extern struct lista my_socket;
	struct sigaction action;
	pthread_attr_t pta;	/**< Struttura degli attributi dei thread che il programma creera' */
	pthread_t tid=0;	/**< Identificatore del thread */
	sigset_t set;		/**< Maschera temporanea per bloccare tutti i segnali */
	channel_t sk_cl;	/**< Canale di comunicazione del server */
	char * path;		/**< Path della socket */
	struct stat buf;
	int err;

	path = NULL; fs = NULL; f_ctrl.fat_ptr = NULL;
	sk_ascolto = 0; num_threads = 0; quitflag = 0;

	/* ========= Gestione dei segnali ========= */
	/* Blocco i segnali */
	if (sigfillset(&set) == -1){
    	errore(__FILE__,__LINE__,"Server: error sigfillset",errno);
    	return EXIT_FAILURE;
	}
	if (pthread_sigmask(SIG_SETMASK, &set, NULL) != 0) {
		errore(__FILE__,__LINE__,"Server: pthread_sigmask error 'sigsetmask'",errno);
		return EXIT_FAILURE;
	}
	/* Installo il nuovo gestore */
	memset(&action, sizeof(action), 0);
	if (sigemptyset(&action.sa_mask) == -1){
    	errore(__FILE__,__LINE__,"Server: error sigemptyset",errno);
    	return EXIT_FAILURE;
	}
	/* Ignoro il SIGPIPE */
	action.sa_handler = SIG_IGN;
	if (sigaction(SIGPIPE, &action, NULL) == -1) {
		errore(__FILE__,__LINE__, "Server: sigaction error 'SIGPIPE'", errno);
		exit(EXIT_FAILURE);
	}
	action.sa_handler = gestore;
	action.sa_flags = 0;
	/* Definisco i segnali che voglio gestire */
	if (sigaction(SIGINT, &action, NULL) < 0){
		errore(__FILE__,__LINE__,"Server: sigaction error 'SIGINT'",errno);
		return EXIT_FAILURE;
	}
	if (sigaction(SIGTERM, &action, NULL) < 0){
    	errore(__FILE__,__LINE__,"Server: sigaction error 'SIGTERM'",errno);
    	return EXIT_FAILURE;;
	}
	/* Installo la nuova maschera dei segnali */
	if (pthread_sigmask(SIG_SETMASK, &action.sa_mask, NULL) != 0) {
		errore(__FILE__,__LINE__,"Server: pthread_sigmask error 'sigsetmask'",errno);
		return EXIT_FAILURE;
	}

	/* ========= Filesystem ========= */
	/* Argomento da riga di comando */
	if (argc<2) {
		fprintf(stderr,"Server: errore non e' stato passato alcun filesystem\n");
		return EXIT_FAILURE;
	}
	/* Esiste il file? */
	if (lstat(argv[1], &buf) < 0) {
		errore(__FILE__,__LINE__, "Server: error filesystem",errno);
		return EXIT_FAILURE;
	}
	/* E' un file regolare? */
	if (!(S_ISREG(buf.st_mode))) {
		fprintf(stderr,"Server: error %s is not a regular file\n",argv[1]);
		return EXIT_FAILURE;
	}
	/* Apro in lettura e scrittura il filesystem */
	if (!(fs = fopen(argv[1], "r+"))) {
		errore(__FILE__,__LINE__, "Server: error on open filesystem",errno);
		return EXIT_FAILURE;
	}
	/* Monto il filesystem */
	if ((err = mount_fat(fs, &f_ctrl)) != 0){ fclose(fs); return err; }

	/* ========= Attributi dei thread ========= */
	if (pthread_attr_init(&pta)!=0) {
		errore(__FILE__,__LINE__,"Server: error pthread_attr_init",errno);
		fclose(fs);
		return EXIT_FAILURE;
	}
	/* I nuovi thread che andro' a creare saranno' gia' nello stato 'DETACHED' */
	if (pthread_attr_setdetachstate(&pta, PTHREAD_CREATE_DETACHED)!=0) {
		errore(__FILE__,__LINE__,"Server: error pthread_attr_setdetachstate",errno);
		fclose(fs); pthread_attr_destroy(&pta);
		return EXIT_FAILURE;
	}

	/* Inizializzo la lista dei thread */
	inizializza (&my_socket);

	/* ========= Socket ========= */
	if (!(path = calloc((strlen(TMP) + strlen(SKTNAME) + 1), sizeof(char)))) {
		errore(__FILE__,__LINE__, "Server: error on allocate memory for 'path'",errno);
		free(f_ctrl.fat_ptr); fclose(fs); pthread_attr_destroy(&pta);
		return EXIT_FAILURE;
	}
	strncpy(path,TMP,strlen(TMP));
	strncat(path,SKTNAME,strlen(SKTNAME));
	if ((lstat(path, &buf) == 0) && (unlink(path) < 0))
		errore(__FILE__,__LINE__,"Server: socket unlink",errno);
	/* Creo la socket di ascolto del server */
	if ((sk_ascolto = createServerChannel(path)) == -1) exit(-1);
	else if (sk_ascolto == SFATENAMETOOLONG) {
		fprintf(stderr,"Error Path Too Long (exceeding UNIX_PATH_MAX)\n");
		free(f_ctrl.fat_ptr); free(path); fclose(fs); pthread_attr_destroy(&pta);
		return EXIT_FAILURE;
	}

	/* ========= Ciclo di ascolto delle richieste ========= */
	while (!quitflag) {
		/* Pongo il server in attesa di nuove connessioni da parte dei client */
		if ((sk_cl = acceptConnection(sk_ascolto)) < 0) break;
		else {
			/* Solo se non ho gia' raggiunto il numero massimo di thread fissato */
			pthread_mutex_lock(&lock);
			while (num_threads == THREAD_MAX)
				pthread_cond_wait(&wait_me, &lock);
			/* Creo un nuovo elemento della lista dei socket.
			 * Il nuovo elemento verra' inserito alla testa della lista */
			num_threads++;
			if (ins_new_elem(sk_cl, &my_socket)) break;
			pthread_mutex_unlock(&lock);
			/* Creo un nuovo thread */
			if (pthread_create(&tid, &pta, thread_servernte,
					&(my_socket.top_elem->sk_cl))) {
				errore(__FILE__,__LINE__, "Server: error pthread_create",errno);
				break;
			}
		}
	}

	/* Aspetto la chiusura di ogni thread aperto */
	pthread_mutex_lock(&lock);
	while (num_threads != 0)
		pthread_cond_wait(&wait_me, &lock);
	pthread_mutex_unlock(&lock);

	/* Chiudo la socket del server */
	closeSocket(sk_ascolto);

	/* Elimino la socket dal filesystem (se esiste) */
	if ((lstat(path, &buf) == 0) && (unlink(path) < 0))
		errore(__FILE__,__LINE__,"Server: socket unlink",errno);

	/* Libero le risorse allocate */
	free(path); free(f_ctrl.fat_ptr); fclose(fs); pthread_attr_destroy(&pta);

	exit(EXIT_FAILURE);
}

/**
 * Il mio gestore dei segnali\n
 * Stampa a monitor il codice del segnale ricevuto se questo
 * e' SIGTERM o SIGINT e mette quitflag a 1, altrimenti se e'
 * un SIGPIPE non fa nulla.
 *
 * \param sig codice del segnale ricevuto
 * */
void gestore(int sig) {
	switch (sig) {
		case SIGTERM:
			write(2, "server: signal 15 detected\n", 27);
			quitflag = 1;
		break;
		case SIGINT:
			write(2, "server: signal 2 detected\n", 26);
			quitflag = 1;
		break;
	}
}
