/**
 *	\file		util.h
 *	\author		Tranchida Giulio, No Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.
 * 
 * Queste funzioni hanno il compito di eseguire il parsing della path.\n
 * Ovvero:
 * 	-	Controllare che il path sia assoluto.
 * 	- 	Scindere la path in "path element".
 *	- 	Controllare che ogni "path element" non ecceda la lunghezza
 * 		consentita dal filesystem MAX_LEN_NAME
 * 	-	Navigare all'interno del filesystem seguendo i singoli "path
 * 		element".\n
 * 		Se uno di questi non esiste o non e' una cartella 
 * 		o un file (a seconda dei casi) l'operazione termina con un
 * 		fallimento (Error Not Such File or Directory).
 */
#ifndef UTIL_H_
#define UTIL_H_

/*
 * codifica azioni possibili per le funzioni 'parsing' e 'find' 
 */
#define	LS			0x00	/**< Codifica delle azioni possibili: Listing della catella. */
#define	MKDIR		0x01	/**< Codifica delle azioni possibili: Creazione di una cartella. */
#define	MKFILE		0x02	/**< Codifica delle azioni possibili: Creazione di un file. */
#define WRFILE		0x03	/**< Codifica delle azioni possibili: Scrittura di un file. */
#define	RDFILE		0x04	/**< Codifica delle azioni possibili: Lettura di un file. */

/** Definizione del tipo Action */
typedef unsigned int Action;

/** 
 * <H3>Parsing</H3> 
 * Questa funzione ha come obbiettivo eseguire il parsing della “path”,
 * controllando che il percorso sia corretto ed esistente, lasciando il
 * file pointer del filesystem “fs” nel punto più opportuno per la
 * funzione chiamante.\n Per ulteriori approfondimenti sulla funzione
 * riferirsi alla relazione sul progetto.
 * 
 * @param fs il FILE pointer che consente di acceddre al
 *           file dove il filesystem e' memorizzato.
 * @param f_ctrl puntatore alla struttura che mantiene tutte 
 *               le informazioni di controllo del filesystem.
 * @param path la stringa che contiene il path assoluto del
 *              file che deve essere creato.
 * @param azione l'azione che deve essere intrapesa dalla funzione,
 *              specializzando il comportamento della funzione a
 * 				seconda delle esigenze.
 * @returns
 *
 *  - EBDP (Error Bad Directory Path) se il path specificato non e' 
 *    assoluto, ossia non inizia con il carattere '/'
 *  - ENTL (Error Name Too Long) uno degli elementi del pathname
 *    supera la lunghezza consentita (MAX_LEN_NAME).
 *  - ENSFD (Error Not Such File or Directory) se un elemento del
 *    pah non siste o non e' una directory.
 *  - ENMSD (Error No More Space on Device) se non ci sono piu' 
 *    blocchi disponibili e quindi ci troviamo in una condizione di
 *    Filesystem Full.
 *  - ERFCD (Error Reading Fat Control Data) se si verifica un errore 
 *    leggendo le strutture di controllo del filesystem Fat.
 *  - EWFCD (Error Writing Fat Control Data) se si verifica un errore 
 *    durante la scrittura delle strutture di controllo del filesystem
 *    Fat.
 *  - ERDB (Error Reading Block Data) se si verifica un errore di
 *    lettura su un blocco della Data Region
 *  - EDAEX (Error File/Directory Already Exists) se si verifica che
 *    il file o la cartella esiste gia'.
 *  - SDLIBERROR se si verifica un altro tipo di errore generico.  
 *  - 0 (zero) in caso di successo
 * */ 
int parsing(FILE *fs, struct fat_ctrl *f_ctrl, char *path, Action azione);

/** 
 * <H3>Find</H3> 
 * Questa funzione ha come obbiettivo eseguire il parsing della “path”,
 * controllando che il percorso sia corretto ed esistente, lasciando il
 * file pointer del filesystem “fs” nel punto più opportuno per la
 * funzione chiamante.\n Per ulteriori approfondimenti sulla funzione
 * riferirsi alla relazione sul progetto.
 * 
 * @param fs il FILE pointer che consente di acceddre al
 *           file dove il filesystem e' memorizzato.
 * @param f_ctrl puntatore alla struttura che mantiene tutte 
 *               le informazioni di controllo del filesystem.
 * @param path la stringa che contiene il path assoluto del
 *              file che deve essere creato.
 * @param azione l'azione che deve essere intrapesa dalla funzione,
 *              specializzando il comportamento della funzione a
 * 				seconda delle esigenze.
 * @returns
 *
 *  - EBDP (Error Bad Directory Path) se il path specificato non e' 
 *    assoluto, ossia non inizia con il carattere '/'
 *  - ENTL (Error Name Too Long) uno degli elementi del pathname
 *    supera la lunghezza consentita (MAX_LEN_NAME).
 *  - ENSFD (Error Not Such File or Directory) se un elemento del
 *    pah non siste o non e' una directory.
 *  - ENMSD (Error No More Space on Device) se non ci sono piu' 
 *    blocchi disponibili e quindi ci troviamo in una condizione di
 *    Filesystem Full.
 *  - ERFCD (Error Reading Fat Control Data) se si verifica un errore 
 *    leggendo le strutture di controllo del filesystem Fat.
 *  - EWFCD (Error Writing Fat Control Data) se si verifica un errore 
 *    durante la scrittura delle strutture di controllo del filesystem
 *    Fat.
 *  - ERDB (Error Reading Block Data) se si verifica un errore di
 *    lettura su un blocco della Data Region
 *  - EDAEX (Error File/Directory Already Exists) se si verifica che
 *    il file o la cartella esiste gia'.
 *  - SDLIBERROR se si verifica un altro tipo di errore generico.  
 *  - 0 (zero) in caso di successo
 * */ 
int find(FILE *fs, struct fat_ctrl *f_ctrl, char *path, Action azione);

#endif /*UTIL_H_*/
