/**
 *	\file		util.c
 *	\author		Tranchida Giulio, No Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.\n\n
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */
 
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>

#include "Include/fat.h"
#include "Include/util.h"
#include "Include/sfaterror.h"

/**
 * <h3>Percorso</h3>
 * Funzione ricorsiva - parsing effettivo della path\n
 * Questa funzione scinde il path in 'path element', richiamandosi
 * ricorsivamente fin tanto che il caso base non e' verificato.\n
 * Caso base: path element == ".", o path element == "/".\n
 * In fase di chiusura la funzione richiama su ogni path element
 * scisso, la funzione find, controllando sempre che il valore di
 * ritorno sia zero.\n Se il valore di ritorno e' diverso da zero,
 * siamo in un caso di errore e la funzione ritorna, chiudendo
 * tutte le sue istanze, l'errore al chiamante.
 * 
 * @param fs il FILE pointer che consente di acceddre al
 *           file dove il filesystem e' memorizzato.
 * @param f_ctrl puntatore alla struttura che mantiene tutte 
 *               le informazioni di controllo del filesystem.
 * @param path la stringa che contiene il path assoluto del
 *              file che deve essere creato.
 * @returns
 *
 *  - EBDP (Error Bad Directory Path) se il path specificato non e' 
 *    assoluto, ossia non inizia con il carattere '/'
 *  - ENTL (Error Name Too Long) uno degli elementi del pathname
 *    supera la lunghezza consentita (MAX_LEN_NAME).
 *  - ENSFD (Error Not Such File or Directory) se un elemento del
 *    pah non siste o non e' una directory.
 *  - ENMSD (Error No More Space on Device) se non ci sono piu' 
 *    blocchi disponibili e quindi ci troviamo in una condizione di
 *    Filesystem Full.
 *  - ERFCD (Error Reading Fat Control Data) se si verifica un errore 
 *    leggendo le strutture di controllo del filesystem Fat.
 *  - EWFCD (Error Writing Fat Control Data) se si verifica un errore 
 *    durante la scrittura delle strutture di controllo del filesystem
 *    Fat.
 *  - ERDB (Error Reading Block Data) se si verifica un errore di
 *    lettura su un blocco della Data Region
 *  - EDAEX (Error File/Directory Already Exists) se si verifica che
 *    il file o la cartella esiste gia'.
 *  - SDLIBERROR se si verifica un altro tipo di errore generico.  
 *  - 0 (zero) in caso di successo
 * */ 
static int percorso(FILE *fs, struct fat_ctrl *f_ctrl, char *path);

int parsing(FILE *fs, struct fat_ctrl *f_ctrl, char *path, Action azione){
	char *dirc, *basec, *bname, *dname;
	int ret_val;
	
	if (!(fs) || !(f_ctrl) || !(path)) return EBDP;
	
	/* Metto il puntatore fs all'inizio della data region */
	fseek(fs, f_ctrl->blk_base, SEEK_SET);

	/* Faccio una copia della path */
	if (!(dirc = strdup(path))) return STDLIBERR;
	/* Elimino l'ultimo 'path element' */
	dname = dirname(dirc);
	ret_val = percorso(fs, f_ctrl, dname);

	free(dirc);
	if (ret_val) return ret_val;
	
	/* Reperisco l'ultimo 'path element' */
	if (!(basec = strdup(path))) return STDLIBERR;
	bname = basename(basec);
	if (strlen(bname)>MAX_LEN_NAME){free(basec); return ENTL;} 

	switch(azione){
		case LS: 	/* ritorna settando 'fs' sull'entry della cartella cercata */		
		case RDFILE:/* ritorna settando 'fs' sull'entry del file cercato */
		case WRFILE:/* ritorna settando 'fs' sull'entry del file cercato */
			ret_val = find(fs, f_ctrl, bname, azione); 
		break;		
		case MKFILE:
		case MKDIR:
			;
		break;
	}
	free(basec);
	return ret_val;
}

static int percorso(FILE *fs, struct fat_ctrl *f_ctrl, char *path){
	char *dirc, *basec, *bname, *dname;
	int ret_val = 0;
	
	if (!(fs) || !(f_ctrl) || !(path) ) return EBDP;
	if (!(dirc = strdup(path))) return STDLIBERR;
	if (!(basec = strdup(path))) return STDLIBERR;
	dname = dirname(dirc);
	bname = basename(basec);
	
	if (strlen(bname) > MAX_LEN_NAME) ret_val = ENTL;
	
	if (!ret_val && strcmp(dname, ".") && strcmp(dname, "/"))
		ret_val = percorso(fs, f_ctrl, dname);

	if (!ret_val && !strcmp(dname, ".")) ret_val = EBDP;
	if (!ret_val) ret_val = find(fs, f_ctrl, bname, LS);

	free(dirc);
	free(basec);
	return ret_val;
}


int find(FILE *fs, struct fat_ctrl *f_ctrl, char *path, Action azione){
	struct dir_entry cart_p;
	int numero_entry, i, j, MAX_ENTRY;
	unsigned int indice_fat, new_index;
	long fs_pos;
	
	if (!(fs) || !(f_ctrl) || !(path)) return EBDP;
	/* Se sono nella radice ritorna */
	if ((azione==LS) && !strcmp(path, "/")) return 0;
	/* Calcolo qual'e' il numero massimo di entry memorizzabile in un
	 *  blocco, in realazione alla dimensione del blocco stesso */
	MAX_ENTRY = f_ctrl->b_sector.block_size/sizeof(struct dir_entry);
	
	fs_pos = ftell(fs);
	/* Leggo la dir_entry "." */
	if (!(fread(&cart_p, sizeof(struct dir_entry), 1, fs))) return ERBD;
	numero_entry = cart_p.len/sizeof(struct dir_entry);
	/* Ricorda l'indice della fat memorizzato nel blocco della entry iniziale '.'
	 * della cartella in cui noi stiamo ciclando  */
	indice_fat = cart_p.index;
	if (indice_fat > f_ctrl->b_sector.num_block) return ERFCD;
	
	/* Relative path navigation "/." (Solo in caso di listing delle cartelle) */
	if ((azione==LS) && !strcmp(cart_p.name, path) && (cart_p.attr == SUB_ENTRY)){
		fseek(fs, fs_pos, SEEK_SET);
		return 0;
	}
	
	for (i=1, j=1; i<numero_entry; i++, j++){		
		if (j==MAX_ENTRY){
			/* Ho letto tutte le entry del blocco.
			 * Devo controllare il blocco DATA REGION successivo nella fat */			
			if (f_ctrl->fat_ptr[indice_fat] == BLOCK_FREE) return ERFCD;
			if (f_ctrl->fat_ptr[indice_fat] == LAST_BLOCK) return ERFCD;				
			/* La posizione successiva di 'fs' e' uguale a blk_base +
			 * la posizione indicata nell'array della FAT * la dimensione dei blocchi. */
			fs_pos = f_ctrl->blk_base + f_ctrl->fat_ptr[indice_fat] * f_ctrl->b_sector.block_size;
			fseek(fs, fs_pos, SEEK_SET);
			/* Aggiorno l'indice della FAT col nuovo valore memorizzato nella locazione stessa della FAT */
			indice_fat = f_ctrl->fat_ptr[indice_fat];
			/* if (indice_fat > f_ctrl->b_sector.num_block) return ERFCD; */
			j=1;
		}
		/* Salvo la posizione dell'entry del file. Utile solo per le azioni WRFILE e RDFILE */
		fs_pos = ftell(fs);
		/* Leggo la dir_entry successiva */
		if (!(fread(&cart_p, sizeof(struct dir_entry), 1, fs))) return ERBD;
		/* Ho trovato il nome che cercavo ed e' una cartella (condizione di uscita) */
		if ((azione==LS) && !strcmp(cart_p.name, path) && (cart_p.attr == SUB_ENTRY)) break;
		/* Ho trovato il nome che cercavo ed e' una cartella ma non avrei dovuto trovarlo!! */
		if ((azione==MKDIR) && !strcmp(cart_p.name, path) && (cart_p.attr ==SUB_ENTRY)) return EDAEX;
		/* Ho trovato il nome che cercavo ed e' un file ma non avrei dovuto trovarlo!! */
		if ((azione==MKFILE) && !strcmp(cart_p.name, path) && (cart_p.attr==FILE_ENTRY)) return EDAEX;
		/* Ho trovato il nome che cercavo ed e' un file (condizione di uscita) */
		if ((azione==WRFILE || azione==RDFILE) && !strcmp(cart_p.name, path) && (cart_p.attr==FILE_ENTRY)) break;
	}

	switch(azione){
		case LS:
			/* Cartella non trovata (Error Not Such File or Directory) */
			if (strcmp(cart_p.name, path)) return ENSFD;
			else if (!strcmp(cart_p.name, path) && (cart_p.attr==FILE_ENTRY)) return ENSFD;
			/* Posiziono il file pointer fs all'inizio del blocco che contiene la prima
			 * porzione della Directory Table relativa alla cartella cercata */
			fs_pos = f_ctrl->blk_base + cart_p.index * f_ctrl->b_sector.block_size;
			fseek(fs, fs_pos, SEEK_SET);
		break;
		case RDFILE:
		case WRFILE:
			/* File non trovato (Error Not Such File or Directory) */
			if (strcmp(cart_p.name, path) || (cart_p.attr!=FILE_ENTRY)) return ENSFD;
			/* Posiziono il file pointer fs all'inizio dell'entry del file su cui devo
			 * andare a scrivere (o a leggere) */
			fseek(fs, fs_pos, SEEK_SET);
		break;
		case MKFILE:
		case MKDIR:
			/* La funzione ritorna settando 'fs' sul primo spazio vuoto del blocco dalla
			 * data region di modo che la write della nuova entry sia nel punto corretto */		
			if (j==MAX_ENTRY){
				/* Cerco un nuovo indice di blocco libero nella FAT */
				for (new_index=0; new_index<f_ctrl->b_sector.num_block; new_index++)
					if (f_ctrl->fat_ptr[new_index] == BLOCK_FREE) break;
				if (new_index>=f_ctrl->b_sector.num_block) return ENMSD;
				/* Imposto il blocco libero trovato della FAT ad LAST_BLOCK. */
				f_ctrl->fat_ptr[new_index] = LAST_BLOCK;
				fs_pos = f_ctrl->blk_base + new_index * f_ctrl->b_sector.block_size;
				/* Aggiorno la locazione della fat puntata dal vecchio indice */
				f_ctrl->fat_ptr[indice_fat]= new_index;
				/* Setto il puntatore 'fs' al nuovo blocco libero della fat */
				fseek(fs, fs_pos, SEEK_SET);
			}		
		break;
	}
	return 0;
}
